import numpy as np
from begonia import Polygon, Point2d
from benches import runtime, runtime_block

N_ITER: int = 1000
N_REPS: int = 10


def polygon_n(n: int):
    theta = np.linspace(0, np.pi, n)
    return Polygon(np.array([np.cos(theta), np.sin(theta)]).T)


runtime_block(
    "begonia.Polygon",
    f"Polygon runtime benchmarks were measured across {N_REPS} repeats with {
        N_ITER} iterations per repeat. A single warm-up run was discarded."
)


def make_benchmark(func_name, label_fmt, *args):
    for n in [10, 100, 1000]:
        @runtime(label_fmt.format(n=n), number=N_ITER, repeat=N_REPS)
        def fn(n=n):
            p = polygon_n(n)
            getattr(p, func_name)(*args)
            return p
        globals()[f"bench_{func_name}_{n}"] = fn


def make_benchmark_polygon(func_name, label_fmt):
    for n in [10, 100, 1000]:
        @runtime(label_fmt.format(n=n), number=N_ITER, repeat=N_REPS)
        def fn(n=n):
            p = polygon_n(n)
            q = polygon_n(n)
            getattr(p, func_name)(q)
            return p
        globals()[f"bench_{func_name}_{n}"] = fn


make_benchmark("to_list", "to_list (n = {n})")
make_benchmark("to_numpy", "to_numpy (n = {n})")
make_benchmark("push", "push (n = {n})", [10., 10.])
make_benchmark("push_point2d", "push_point2d (n = {n})", Point2d(10., 10.))
make_benchmark("add_scalar_inplace", "add_scalar_inplace (n = {n})", 10)
make_benchmark("sub_scalar_inplace", "sub_scalar_inplace (n = {n})", 10)
make_benchmark("mul_scalar_inplace", "mul_scalar_inplace (n = {n})", 10)
make_benchmark("div_scalar_inplace", "div_scalar_inplace (n = {n})", 10)
make_benchmark("add_scalar", "add_scalar (n = {n})", 10)
make_benchmark("sub_scalar", "sub_scalar (n = {n})", 10)
make_benchmark("mul_scalar", "mul_scalar (n = {n})", 10)
make_benchmark("div_scalar", "div_scalar (n = {n})", 10)
make_benchmark("add_point_inplace", "add_point_inplace (n = {n})", [5, 10])
make_benchmark("add_point2d_inplace",
               "add_point2d_inplace (n = {n})", Point2d(5, 10))
make_benchmark("sub_point_inplace", "sub_point_inplace (n = {n})", [5, 10])
make_benchmark("sub_point2d_inplace",
               "sub_point2d_inplace (n = {n})", Point2d(5, 10))
make_benchmark("mul_point_inplace", "mul_point_inplace (n = {n})", [5, 10])
make_benchmark("mul_point2d_inplace",
               "mul_point2d_inplace (n = {n})", Point2d(5, 10))
make_benchmark("div_point_inplace", "div_point_inplace (n = {n})", [5, 10])
make_benchmark("div_point2d_inplace",
               "div_point2d_inplace (n = {n})", Point2d(5, 10))
make_benchmark("add_point", "add_point (n = {n})", [5, 10])
make_benchmark("add_point2d", "add_point2d (n = {n})", Point2d(5, 10))
make_benchmark("sub_point", "sub_point (n = {n})", [5, 10])
make_benchmark("sub_point2d", "sub_point2d (n = {n})", Point2d(5, 10))
make_benchmark("mul_point", "mul_point (n = {n})", [5, 10])
make_benchmark("mul_point2d", "mul_point2d (n = {n})", Point2d(5, 10))
make_benchmark("div_point", "div_point (n = {n})", [5, 10])
make_benchmark("div_point2d", "div_point2d (n = {n})", Point2d(5, 10))
make_benchmark("dedup", "dedup (n = {n})")
make_benchmark("dedup_inplace", "dedup_inplace (n = {n})")
make_benchmark("order", "order (n = {n})")
make_benchmark("order_inplace", "order_inplace (n = {n})")
make_benchmark("resample", "resample (n = {n})", 5)
make_benchmark("resample_inplace", "resample_inplace (n = {n})", 5)
make_benchmark("encloses_point", "encloses_point (n = {n})", [1., 1.])
make_benchmark("encloses_point2d",
               "encloses_point2d (n = {n})", Point2d(1., 1.))
make_benchmark("distance_to_point_center",
               "distance_to_point_center (n = {n})", [1., 1.])
make_benchmark("distance_to_point2d_center",
               "distance_to_point2d_center (n = {n})", Point2d(1., 1.))
make_benchmark("distance_to_point_centroid",
               "distance_to_point_centroid (n = {n})", [1., 1.])
make_benchmark("distance_to_point2d_centroid",
               "distance_to_point2d_centroid (n = {n})", Point2d(1., 1.))
make_benchmark("distance_to_point_vertex",
               "distance_to_point_vertex (n = {n})", [1., 1.])
make_benchmark("distance_to_point2d_vertex",
               "distance_to_point2d_vertex (n = {n})", Point2d(1., 1.))
make_benchmark("distance_to_point_edge",
               "distance_to_point_edge (n = {n})", [1., 1.])
make_benchmark("distance_to_point2d_edge",
               "distance_to_point2d_edge (n = {n})", Point2d(1., 1.))
make_benchmark("convex_hull", "convex_hull (n = {n})")
make_benchmark("center", "center (n = {n})")
make_benchmark("centroid", "centroid (n = {n})")
make_benchmark("area", "area (n = {n})")
make_benchmark("area_bbox", "area_bbox (n = {n})")
make_benchmark("area_convex", "area_convex (n = {n})")
make_benchmark("perimeter", "perimeter (n = {n})")
make_benchmark("elongation", "elongation (n = {n})")
make_benchmark("thread_length", "thread_length (n = {n})")
make_benchmark("solidity", "solidity (n = {n})")
make_benchmark("extent", "extent (n = {n})")
make_benchmark("form_factor", "form_factor (n = {n})")
make_benchmark("equivalent_diameter", "equivalent_diameter (n = {n})")
make_benchmark("eccentricity", "eccentricity (n = {n})")
make_benchmark("major_axis_length", "major_axis_length (n = {n})")
make_benchmark("minor_axis_length", "minor_axis_length (n = {n})")
make_benchmark("min_radius", "min_radius (n = {n})")
make_benchmark("max_radius", "max_radius (n = {n})")
make_benchmark("mean_radius", "mean_radius (n = {n})")
make_benchmark("min_feret", "min_feret (n = {n})")
make_benchmark("max_feret", "max_feret (n = {n})")

make_benchmark_polygon("add", "add (n = {n})")
make_benchmark_polygon("sub", "sub (n = {n})")
make_benchmark_polygon("mul", "mul (n = {n})")
make_benchmark_polygon("div", "div (n = {n})")
make_benchmark_polygon("d_l1", "d_l1 (n = {n})")
make_benchmark_polygon("d_l2", "d_l2 (n = {n})")
make_benchmark_polygon("d_chebyshev", "d_chebyshev (n = {n})")
make_benchmark_polygon("d_cosine", "d_cosine (n = {n})")
make_benchmark_polygon("distance_to_polygon_center",
                       "distance_to_polygon_center (n = {n})")
make_benchmark_polygon("distance_to_polygon_centroid",
                       "distance_to_polygon_centroid (n = {n})")
make_benchmark_polygon("distance_to_polygon_vertex",
                       "distance_to_polygon_vertex (n = {n})")
make_benchmark_polygon("distance_to_polygon_edge",
                       "distance_to_polygon_edge (n = {n})")
make_benchmark_polygon("align_to",
                       "align_to (n = {n})")


def run_all_benchmarks():
    for name, fn in globals().items():
        if name.startswith("bench_") and callable(fn):
            fn()


if __name__ == "__main__":
    run_all_benchmarks()



import timeit
import functools


def runtime_block(header: str, description: str):
    print(
        f"## {header}\n\n" +
        f"{description}\n\n" +
        "| Bench | Minimum (s) | Maximum (s) | Average (s) |\n" +
        "| ----- | ----------- | ----------- | ----------- |"
    )


def runtime(name: str, number: int = 1000, repeat: int = 5):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            # Warm-up run
            func(*args, **kwargs)

            def func_to_benchmark():
                func(*args, **kwargs)

            times = timeit.repeat(
                func_to_benchmark,
                number=number,
                repeat=repeat
            )

            min_time = min(times) / number
            max_time = max(times) / number
            avg_time = (sum(times) / len(times)) / number

            print(f"| {name} | {min_time:.3e} | {
                  max_time:.3e} | {avg_time:.3e} |")

            return min_time, avg_time
        return wrapper
    return decorator



import pytest
import numpy as np
from begonia import Polygon, Point2d


def polygon_unwrapped():
    return Polygon([[0, 0], [1, 0], [1, 1], [0, 1]])


def polygon_unwrapped_negative():
    return Polygon([[0, 0], [-1, 0], [-1, -1], [0, -1]])


def polygon_wrapped():
    return Polygon([[0, 0], [1, 0], [1, 1], [0, 1], [0, 0]])


def unit_circle():
    points = []
    for i in np.hstack([
        np.linspace(0, np.pi / 2., 250),
        np.linspace(np.pi / 2., np.pi, 250),
        np.linspace(np.pi, 3 * np.pi / 2., 250),
        np.linspace(3 * np.pi / 2., 2 * np.pi, 250),
    ]):
        x = np.cos(i)
        y = np.sin(i)
        points.append([x, y])
    return Polygon(points)


class TestPolygon:
    def test_new_success(self):
        p = polygon_unwrapped()
        assert isinstance(p.xy, list)
        assert isinstance(p.xy[0], list)
        assert len(p.xy) == 4
        assert isinstance(p, Polygon)

    def test_new_invalid_type(self):
        with pytest.raises(TypeError):
            Polygon([0, "A"])

    def test_new_invalid_size(self):
        with pytest.raises(ValueError):
            Polygon([[0, 0], [0, 1]])

    def test_eq(self):
        p = polygon_unwrapped()
        q = polygon_wrapped()
        assert hasattr(p, "__eq__")
        assert p == p
        assert p != q

    def test_approximate_eq(self):
        p = polygon_unwrapped()
        q = polygon_unwrapped()
        assert p.eq(q, eps=1e-20)

        q = Polygon([[1e-6, 0], [1, 0], [1, 1], [0, 1]])
        assert p.eq(q, eps=1e-5)
        assert not p.eq(q, eps=1e-7)

    def test_len(self):
        p = polygon_unwrapped()
        assert len(p) == 4

    def test_iter(self):
        p = polygon_unwrapped()
        for i in p:
            assert isinstance(i, list)
            assert len(i) == 2
            assert isinstance(i[0], float)
            assert isinstance(i[1], float)

    def test_to_list(self):
        p = polygon_unwrapped()
        assert isinstance(p.to_list(), list)

    def test_to_numpy(self):
        p = polygon_unwrapped()
        assert isinstance(p.to_numpy(), np.ndarray)

    def test_push(self):
        p = polygon_unwrapped()
        p.push([10., 10.])
        assert p.xy[-1] == [10., 10.]

    def test_push_point2d(self):
        p = polygon_unwrapped()
        a = Point2d(10., 10.)
        p.push_point2d(a)
        assert p.xy[-1] == [10., 10.]

    def test_add_scalar_inplace(self):
        p = polygon_unwrapped()
        q = p.to_numpy()
        p.add_scalar_inplace(10)
        for i, j in zip(q.flatten(), p.to_numpy().flatten()):
            assert i + 10 == j

    def test_sub_scalar_inplace(self):
        p = polygon_unwrapped()
        q = p.to_numpy()
        p.sub_scalar_inplace(10)
        for i, j in zip(q.flatten(), p.to_numpy().flatten()):
            assert i - 10 == j

    def test_mul_scalar_inplace(self):
        p = polygon_unwrapped()
        q = p.to_numpy()
        p.mul_scalar_inplace(10)
        for i, j in zip(q.flatten(), p.to_numpy().flatten()):
            assert i * 10 == j

    def test_div_scalar_inplace(self):
        p = polygon_unwrapped()
        q = p.to_numpy()
        p.div_scalar_inplace(10)
        for i, j in zip(q.flatten(), p.to_numpy().flatten()):
            assert i / 10 == j

    def test_add_scalar(self):
        p = polygon_unwrapped()
        q = p.to_numpy()
        p = p.add_scalar(10)
        for i, j in zip(q.flatten(), p.to_numpy().flatten()):
            assert i + 10 == j

    def test_sub_scalar(self):
        p = polygon_unwrapped()
        q = p.to_numpy()
        p = p.sub_scalar(10)
        for i, j in zip(q.flatten(), p.to_numpy().flatten()):
            assert i - 10 == j

    def test_mul_scalar(self):
        p = polygon_unwrapped()
        q = p.to_numpy()
        p = p.mul_scalar(10)
        for i, j in zip(q.flatten(), p.to_numpy().flatten()):
            assert i * 10 == j

    def test_div_scalar(self):
        p = polygon_unwrapped()
        q = p.to_numpy()
        p = p.div_scalar(10)
        for i, j in zip(q.flatten(), p.to_numpy().flatten()):
            assert i / 10 == j

    def test_add_point_inplace(self):
        p = polygon_unwrapped()
        q = p.to_numpy()
        p.add_point_inplace([5, 10])
        for i, j in zip(q, p):
            assert i[0] + 5 == j[0]
            assert i[1] + 10 == j[1]

    def test_add_point2d_inplace(self):
        p = polygon_unwrapped()
        q = p.to_numpy()
        p.add_point2d_inplace(Point2d(5, 10))
        for i, j in zip(q, p):
            assert i[0] + 5 == j[0]
            assert i[1] + 10 == j[1]

    def test_sub_point_inplace(self):
        p = polygon_unwrapped()
        q = p.to_numpy()
        p.sub_point_inplace([5, 10])
        for i, j in zip(q, p):
            assert i[0] - 5 == j[0]
            assert i[1] - 10 == j[1]

    def test_sub_point2d_inplace(self):
        p = polygon_unwrapped()
        q = p.to_numpy()
        p.sub_point2d_inplace(Point2d(5, 10))
        for i, j in zip(q, p):
            assert i[0] - 5 == j[0]
            assert i[1] - 10 == j[1]

    def test_mul_point_inplace(self):
        p = polygon_unwrapped()
        q = p.to_numpy()
        p.mul_point_inplace([5, 10])
        for i, j in zip(q, p):
            assert i[0] * 5 == j[0]
            assert i[1] * 10 == j[1]

    def test_mul_point2d_inplace(self):
        p = polygon_unwrapped()
        q = p.to_numpy()
        p.mul_point2d_inplace(Point2d(5, 10))
        for i, j in zip(q, p):
            assert i[0] * 5 == j[0]
            assert i[1] * 10 == j[1]

    def test_div_point_inplace(self):
        p = polygon_unwrapped()
        q = p.to_numpy()
        p.div_point_inplace([5, 10])
        for i, j in zip(q, p):
            assert i[0] / 5 == j[0]
            assert i[1] / 10 == j[1]

    def test_div_point2d_inplace(self):
        p = polygon_unwrapped()
        q = p.to_numpy()
        p.div_point2d_inplace(Point2d(5, 10))
        for i, j in zip(q, p):
            assert i[0] / 5 == j[0]
            assert i[1] / 10 == j[1]

    def test_add_point(self):
        p = polygon_unwrapped()
        q = p.to_numpy()
        p = p.add_point([5, 10])
        for i, j in zip(q, p):
            assert i[0] + 5 == j[0]
            assert i[1] + 10 == j[1]

    def test_add_point2d(self):
        p = polygon_unwrapped()
        q = p.to_numpy()
        p = p.add_point2d(Point2d(5, 10))
        for i, j in zip(q, p):
            assert i[0] + 5 == j[0]
            assert i[1] + 10 == j[1]

    def test_sub_point(self):
        p = polygon_unwrapped()
        q = p.to_numpy()
        p = p.sub_point([5, 10])
        for i, j in zip(q, p):
            assert i[0] - 5 == j[0]
            assert i[1] - 10 == j[1]

    def test_sub_point2d(self):
        p = polygon_unwrapped()
        q = p.to_numpy()
        p = p.sub_point2d(Point2d(5, 10))
        for i, j in zip(q, p):
            assert i[0] - 5 == j[0]
            assert i[1] - 10 == j[1]

    def test_mul_point(self):
        p = polygon_unwrapped()
        q = p.to_numpy()
        p = p.mul_point([5, 10])
        for i, j in zip(q, p):
            assert i[0] * 5 == j[0]
            assert i[1] * 10 == j[1]

    def test_mul_point2d(self):
        p = polygon_unwrapped()
        q = p.to_numpy()
        p = p.mul_point2d(Point2d(5, 10))
        for i, j in zip(q, p):
            assert i[0] * 5 == j[0]
            assert i[1] * 10 == j[1]

    def test_div_point(self):
        p = polygon_unwrapped()
        q = p.to_numpy()
        p = p.div_point([5, 10])
        for i, j in zip(q, p):
            assert i[0] / 5 == j[0]
            assert i[1] / 10 == j[1]

    def test_div_point2d(self):
        p = polygon_unwrapped()
        q = p.to_numpy()
        p = p.div_point2d(Point2d(5, 10))
        for i, j in zip(q, p):
            assert i[0] / 5 == j[0]
            assert i[1] / 10 == j[1]

    def test_add_inplace(self):
        p = polygon_unwrapped()
        q = polygon_unwrapped()
        p.add_inplace(q)
        assert p == Polygon([[0, 0], [2, 0], [2, 2], [0, 2]])

    def test_sub_inplace(self):
        p = polygon_unwrapped()
        q = polygon_unwrapped()
        p.sub_inplace(q)
        assert p == Polygon([[0, 0], [0, 0], [0, 0], [0, 0]])

    def test_mul_inplace(self):
        p = polygon_unwrapped()
        q = polygon_unwrapped()
        p.mul_inplace(q)
        assert p == q

    def test_div_inplace(self):
        p = polygon_unwrapped()
        q = polygon_unwrapped()
        p.add_scalar_inplace(2)
        q.add_scalar_inplace(1)
        z = p.to_numpy()
        p.div_inplace(q)
        assert np.allclose(p.to_numpy(), z / q.to_numpy())

    def test_add(self):
        p = q = polygon_unwrapped()
        p = p.add(q)
        assert p == Polygon([[0, 0], [2, 0], [2, 2], [0, 2]])

    def test_sub(self):
        p = q = polygon_unwrapped()
        p = p.sub(q)
        assert p == Polygon([[0, 0], [0, 0], [0, 0], [0, 0]])

    def test_mul(self):
        p = q = polygon_unwrapped()
        p = p.mul(q)
        assert p == q

    def test_div(self):
        p = q = polygon_unwrapped()
        p.add_scalar_inplace(2)
        q.add_scalar_inplace(1)
        z = p.div(q)
        assert np.allclose(z.to_numpy(), p.to_numpy() / q.to_numpy())

    def test_center(self):
        p = polygon_unwrapped()
        assert p.center() == Point2d(0.5, 0.5)

        q = polygon_wrapped()
        assert q.center() == Point2d(0.5, 0.5)

        z = polygon_unwrapped_negative()
        assert z.center() == Point2d(-0.5, -0.5)

    def test_centroid(self):
        p = polygon_unwrapped()
        x = p.centroid()
        assert p.centroid() == Point2d(0.5, 0.5)

        q = polygon_wrapped()
        assert q.centroid() == Point2d(0.5, 0.5)

        z = polygon_unwrapped_negative()
        assert z.centroid() == Point2d(-0.5, -0.5)

    def test_d_l1(self):
        p = q = polygon_unwrapped()
        q = q.add(q)
        p = p.d_l1(q)
        assert p == 4

    def test_d_l2(self):
        p = q = polygon_unwrapped()
        q = q.add(q)
        p = p.d_l2(q)
        assert p == 2

    def test_d_chebyshev(self):
        p = q = polygon_unwrapped()
        q = q.add(q)
        p = p.d_chebyshev(q)
        assert p == 1

    def test_d_cosine(self):
        p = q = polygon_unwrapped()
        q = q.add(q)
        d = p.d_cosine(q)
        assert d == 0.0

    def test_dedup(self):
        p = polygon_unwrapped()
        q = polygon_wrapped()
        q = q.dedup()
        assert q == p

    def test_dedup_inplace(self):
        p = polygon_unwrapped()
        q = polygon_wrapped()
        q.dedup_inplace()
        assert q == p

    def test_dedup_unstable(self):
        p = polygon_unwrapped()
        q = polygon_wrapped()
        q = q.dedup()
        assert len(q) == 4
        for i in q:
            count = sum([i == j for j in p])
            assert count == 1

    def test_dedup_unstable_inplace(self):
        p = polygon_unwrapped()
        q = polygon_wrapped()
        q.dedup_inplace()
        assert len(q) == 4
        for i in q:
            count = sum([i == j for j in p])
            assert count == 1

    def test_order(self):
        p = Polygon([[0, 1], [1, 0], [0, 0], [1, 1]])
        p = p.order()
        assert p.to_list() == [[0., 1.], [1., 1.], [1., 0.], [0., 0.]]

    def test_order_inplace(self):
        p = Polygon([[0, 1], [1, 0], [0, 0], [1, 1]])
        p.order_inplace()
        assert p.to_list() == [[0., 1.], [1., 1.], [1., 0.], [0., 0.]]

    def test_resample(self):
        p = polygon_unwrapped()
        q = p.resample(5)
        assert q == p

    def test_resample_inplace(self):
        p = polygon_unwrapped()
        q = polygon_unwrapped()
        p.resample_inplace(5)
        assert p == q

    def test_encloses_point(self):
        p = polygon_unwrapped()
        p1 = [0.5, 0.5]
        assert p.encloses_point(p1)
        assert p.encloses_point(np.array(p1))

        p2 = [-10., 10]
        assert not p.encloses_point(p2)
        assert not p.encloses_point(np.array(p2))

        p3 = [0., 0.]
        assert p.encloses_point(p3)
        assert p.encloses_point(np.array(p3))

        triangle = Polygon([[0., 0.], [1., 1.], [2., 0.]])
        assert triangle.encloses_point(p1)
        assert triangle.encloses_point(np.array(p1))

        l_shape = Polygon([
            [0., 0.], [1., 0.], [2., 0.], [2., 2.], [1., 2.], [1., 1.], [0., 1]])
        p4 = [1.75, 1.75]
        p5 = [0.1, 0.1]
        assert l_shape.encloses_point(p4)
        assert l_shape.encloses_point(np.array(p4))
        assert l_shape.encloses_point(p5)
        assert l_shape.encloses_point(np.array(p5))

    def test_encloses_point2d(self):
        p = polygon_unwrapped()
        p1 = Point2d(0.5, 0.5)
        assert p.encloses_point2d(p1)

        p2 = Point2d(-10., 10)
        assert not p.encloses_point2d(p2)

        p3 = Point2d(0., 0.)
        assert p.encloses_point2d(p3)

        triangle = Polygon([[0., 0.], [1., 1.], [2., 0.]])
        assert triangle.encloses_point2d(p1)

        l_shape = Polygon([
            [0., 0.], [1., 0.], [2., 0.], [2., 2.], [1., 2.], [1., 1.], [0., 1]])
        p4 = Point2d(1.75, 1.75)
        p5 = Point2d(0.1, 0.1)
        assert l_shape.encloses_point2d(p4)
        assert l_shape.encloses_point2d(p5)

    def test_distance_to_point_center(self):
        p = polygon_unwrapped()
        p1 = [0.5, 0.5]
        assert p.distance_to_point_center(p1) == 0.0
        assert p.distance_to_point_center(np.array(p1)) == 0.0

        p2 = [10, 12]
        assert p.distance_to_point_center(p2) == np.sqrt(9.5**2 + 11.5**2)
        assert p.distance_to_point_center(
            np.array(p2)) == np.sqrt(9.5**2 + 11.5**2)

    def test_distance_to_point2d_center(self):
        p = polygon_unwrapped()
        p1 = Point2d(0.5, 0.5)
        assert p.distance_to_point2d_center(p1) == 0.0

        p2 = Point2d(10, 12)
        assert p.distance_to_point2d_center(p2) == np.sqrt(9.5**2 + 11.5**2)

    def test_distance_to_point_centroid(self):
        p = polygon_unwrapped()
        p1 = [0.5, 0.5]
        assert p.distance_to_point_centroid(p1) == 0.0
        assert p.distance_to_point_centroid(np.array(p1)) == 0.0

        p2 = [10, 12]
        assert p.distance_to_point_centroid(p2) == np.sqrt(9.5**2 + 11.5**2)
        assert p.distance_to_point_centroid(
            np.array(p2)) == np.sqrt(9.5**2 + 11.5**2)

    def test_distance_to_point2d_centroid(self):
        p = polygon_unwrapped()
        p1 = Point2d(0.5, 0.5)
        assert p.distance_to_point2d_centroid(p1) == 0.0

        p2 = Point2d(10, 12)
        assert p.distance_to_point2d_centroid(p2) == np.sqrt(9.5**2 + 11.5**2)

    def test_distance_to_point_vertex(self):
        p = polygon_unwrapped()
        p1 = [0.5, 0.5]
        assert p.distance_to_point_vertex(p1) == np.sqrt(2*(0.5**2))
        assert p.distance_to_point_vertex(np.array(p1)) == np.sqrt(2*(0.5**2))

        p2 = [0, 10]
        assert p.distance_to_point_vertex(p2) == 9.0
        assert p.distance_to_point_vertex(np.array(p2)) == 9.0

    def test_distance_to_point2d_vertex(self):
        p = polygon_unwrapped()
        p1 = Point2d(0.5, 0.5)
        assert p.distance_to_point2d_vertex(p1) == np.sqrt(2*(0.5**2))

        p2 = Point2d(0, 10)
        assert p.distance_to_point2d_vertex(p2) == 9.0

    def test_distance_to_point_edge(self):
        p = polygon_unwrapped()
        p1 = [0.5, 0.5]
        assert p.distance_to_point_edge(p1) == 0.5
        assert p.distance_to_point_edge(np.array(p1)) == 0.5

        p2 = [0.5, 10]
        assert p.distance_to_point_edge(p2) == 9.0
        assert p.distance_to_point_edge(np.array(p2)) == 9.0

    def test_distance_to_point2d_edge(self):
        p = polygon_unwrapped()
        p1 = Point2d(0.5, 0.5)
        assert p.distance_to_point2d_edge(p1) == 0.5

        p2 = Point2d(0.5, 10)
        assert p.distance_to_point2d_edge(p2) == 9.0

    def test_distance_to_polygon_center(self):
        p = polygon_unwrapped()
        q = p.add_point([9., 0.])
        assert p.distance_to_polygon_center(q) == 9.

    def test_distance_to_polygon_centroid(self):
        p = polygon_unwrapped()
        q = p.add_point([9., 0.])
        assert p.distance_to_polygon_centroid(q) == 9.

    def test_distance_to_polygon_vertex(self):
        p = polygon_unwrapped()
        q = p.add_point([9., 1.])
        assert p.distance_to_polygon_vertex(q) == 8.

    def test_distance_to_polygon_edge(self):
        p = polygon_unwrapped()
        q = p.add_point([9., 0.])
        assert p.distance_to_polygon_edge(q) == 8.

    def test_convex_hull(self):
        p = polygon_unwrapped()
        assert p.convex_hull().to_list() == [[0., 0.], [
            0., 1.], [1., 1.], [1., 0.]]

        q = Polygon([[0., 0.], [0., 1.], [0.5, 0.9], [1., 1.], [1., 0.]])
        assert q.convex_hull().to_list() == [[0., 0.], [
            0., 1.], [1., 1.], [1., 0.]]

    def test_align_to(self):
        p = polygon_unwrapped()
        q = Polygon([[-0.5, 0], [0.5, 0], [0.5, 1.0], [-0.5, 1.0]])
        q = q.align_to(p)
        assert q == p

        d = np.sqrt(2*(0.5**2))
        s = Polygon([[-d, 0.], [0., -d], [d, 0.], [0., d]])
        s = s.align_to(p)
        assert s.eq(p, eps=1e-15)

        z = polygon_wrapped()
        with pytest.raises(ValueError):
            z = z.align_to(p)

    def test_area(self):
        p = polygon_unwrapped()
        q = polygon_wrapped()
        z = unit_circle()
        assert p.area() == 1.
        assert q.area() == 1.
        assert np.abs(z.area() - np.pi) < 1e-4

    def test_area_bbox(self):
        p = polygon_unwrapped()
        q = polygon_wrapped()
        z = unit_circle()
        assert p.area_bbox() == 1.
        assert q.area_bbox() == 1.
        assert z.area_bbox() == 4.

    def test_area_convex(self):
        p = polygon_unwrapped()
        q = polygon_wrapped()
        z = unit_circle()
        assert p.area_convex() == 1.
        assert q.area_convex() == 1.
        assert z.area_convex() < 2. * np.pi

    def test_perimeter(self):
        p = polygon_unwrapped()
        q = polygon_wrapped()
        z = unit_circle()
        assert p.perimeter() == 4.
        assert q.perimeter() == 4.
        assert np.abs(z.perimeter() - 2 * np.pi) < 1e-4

    def test_elongation(self):
        p = polygon_unwrapped()
        q = polygon_wrapped()
        z = unit_circle()
        assert p.elongation() == 1.
        assert q.elongation() == 1.
        assert z.elongation() == 1.

    def test_thread_length(self):
        p = polygon_unwrapped()
        q = polygon_wrapped()
        z = unit_circle()
        assert p.thread_length() == 1.
        assert q.thread_length() == 1.
        assert np.abs(z.thread_length() - (2 * np.pi) / 4.) < 1e-4

    def test_solidity(self):
        p = polygon_unwrapped()
        q = polygon_wrapped()
        z = unit_circle()
        assert p.solidity() == 1.
        assert q.solidity() == 1.
        assert z.solidity() == z.area() / z.area_convex()

    def test_extent(self):
        p = polygon_unwrapped()
        q = polygon_wrapped()
        z = unit_circle()
        assert p.extent() == 1.
        assert q.extent() == 1.
        assert z.extent() == z.area() / z.area_bbox()

    def test_form_factor(self):
        p = polygon_unwrapped()
        q = polygon_wrapped()
        z = unit_circle()
        assert p.form_factor() == 4. * np.pi / 16.
        assert q.form_factor() == 4. * np.pi / 16.
        assert z.form_factor() == 4. * np.pi * z.area() / z.perimeter()**2

    def test_equivalent_diameter(self):
        p = polygon_unwrapped()
        q = polygon_wrapped()
        z = unit_circle()
        assert p.equivalent_diameter() == np.sqrt(4. / np.pi)
        assert q.equivalent_diameter() == np.sqrt(4. / np.pi)
        assert z.equivalent_diameter() == np.sqrt(4. * z.area() / np.pi)

    def test_eccentricity(self):
        p = polygon_unwrapped()
        q = polygon_wrapped()
        z = unit_circle()
        assert np.abs(p.eccentricity() - 0.) < 1e-7
        assert np.abs(q.eccentricity() - 0.) < 1e-7
        assert np.abs(z.eccentricity() - 0.) < 1e-5

    def test_major_axis_length(self):
        p = polygon_unwrapped()
        q = polygon_wrapped()
        z = unit_circle()
        assert np.abs(p.major_axis_length() - np.sqrt(2.)) < 1e-8
        assert np.abs(q.major_axis_length() - np.sqrt(2.)) < 1e-8
        assert np.abs(z.major_axis_length() - 2.) < 1e-8

    def test_minor_axis_length(self):
        p = polygon_unwrapped()
        q = polygon_wrapped()
        z = unit_circle()
        assert np.abs(p.major_axis_length() - np.sqrt(2.)) < 1e-8
        assert np.abs(q.major_axis_length() - np.sqrt(2.)) < 1e-8
        assert np.abs(z.major_axis_length() - 2.) < 1e-8

    def test_min_radius(self):
        p = polygon_unwrapped()
        q = polygon_wrapped()
        z = unit_circle()
        assert p.min_radius() == 0.5
        assert q.min_radius() == 0.5
        assert np.abs(z.min_radius() - 1.) < 1e-5

    def test_max_radius(self):
        p = polygon_unwrapped()
        q = polygon_wrapped()
        z = unit_circle()
        assert p.max_radius() == np.sqrt(2.) / 2.
        assert q.max_radius() == np.sqrt(2.) / 2.
        assert np.abs(z.max_radius() - 1.) < 1e-15

    def test_mean_radius(self):
        p = polygon_unwrapped()
        q = polygon_wrapped()
        z = unit_circle()
        print((0.5 + np.sqrt(2.) / 2.) / 2.)
        assert p.mean_radius() == np.sqrt(2.) / 2.
        assert q.mean_radius() == np.sqrt(2.) / 2.
        assert np.abs(z.mean_radius() - 1.) < 1e-15

    def test_min_feret(self):
        p = polygon_unwrapped()
        q = polygon_wrapped()
        z = unit_circle()
        assert p.min_feret() == 1.
        assert q.min_feret() == 1.
        assert np.abs(z.min_feret() - 2.) < 1e-5

    def test_max_feret(self):
        p = polygon_unwrapped()
        q = polygon_wrapped()
        z = unit_circle()
        assert np.abs(p.max_feret() - np.sqrt(2.)) < 1e-9
        assert np.abs(q.max_feret() - np.sqrt(2.)) < 1e-9
        assert np.abs(z.max_feret() - 2.) < 1e-15



import pytest
import numpy as np
from begonia import Point2d


class TestPoint2d:
    def test_new_success(self):
        p = Point2d(10, 10)
        assert p.x == 10
        assert p.y == 10

    def test_new_invalid_type(self):
        with pytest.raises(TypeError):
            Point2d("A", "B")

    def test_to_list(self):
        p = Point2d(10, 10)
        assert isinstance(p.to_list(), list)

    def test_to_numpy(self):
        p = Point2d(10, 10)
        assert isinstance(p.to_numpy(), np.ndarray)

    def test_eq(self):
        assert Point2d(1, 2) == Point2d(1, 2)

    def test_add(self):
        p = Point2d(4, 2)
        q = Point2d(2, 4)
        assert p.add(q) == Point2d(6, 6)
        assert q.add(p) == Point2d(6, 6)

    def test_sub(self):
        p = Point2d(4, 2)
        q = Point2d(2, 4)
        assert p.sub(q) == Point2d(2, -2)
        assert q.sub(p) == Point2d(-2, 2)

    def test_mul(self):
        p = Point2d(4, 2)
        q = Point2d(2, 4)
        assert p.mul(q) == Point2d(8, 8)
        assert q.mul(p) == Point2d(8, 8)

    def test_div(self):
        p = Point2d(4, 2)
        q = Point2d(2, 4)
        assert p.div(q) == Point2d(2, 0.5)
        assert q.div(p) == Point2d(0.5, 2)

    def test_d_l1(self):
        p = Point2d(2, 4)
        q = Point2d(4, 2)
        assert p.d_l1(q) == 4.
        assert q.d_l1(p) == 4.

    def test_d_l2(self):
        p = Point2d(2, 4)
        q = Point2d(4, 2)
        assert p.d_l2(q) == 8.**(1./2.)
        assert q.d_l2(p) == 8.**(1./2.)

    def test_d_chebyshev(self):
        p = Point2d(2, 4)
        q = Point2d(4, 2)
        assert p.d_chebyshev(q) == 2.
        assert q.d_chebyshev(p) == 2.

    def test_d_cosine(self):
        p = Point2d(2, 4)
        q = Point2d(4, 2)
        assert np.isclose(p.d_cosine(q), 1. - 16. / 20.)
        assert np.isclose(q.d_cosine(p), 1. - 16. / 20.)

    def test_interp(self):
        p = Point2d(2, 4)
        q = Point2d(4, 2)
        assert p.interp(q, 0.0) == p
        assert p.interp(q, 1.0) == q
        assert p.interp(q, 0.5) == Point2d(3, 3)
